    #include "mbed.h"
    #include "TLV320.h"
    #include "SDHCCard.h"
    #include "Si4735.h"
    #include "TextLCD.h"
    #include "Selection.h"
    #include "RPG.h"
    #include "Menu.h"
    #include "Navigator.h"
    #include <vector>
    #include <string>
    
 using namespace std;

    SDHCCard sd (p11, p12, p13, p20, "sd"); 
    TLV320 audio(p9, p10, 52, p5, p6, p7, p8, p29);         //TLV320 object (Stereo Audio CODEC)
    // constructor for the Si4735 AM/FM radio
    // Si4735(PinName sda, PinName scl, PinName RST_)
    Si4735 radio(p9, p10, p25);
    TextLCD lcd(p19, p18, p14, p15, p16, p17);      // rs, e, d4-d7
    RPG rpg(p26,p24,p28);                           // rotary pulse generator
    Menu rootMenu("root"); 
    Navigator navigator(&rootMenu, rpg, &lcd);
 
      
    //mbed NXP LPC1768: Any of the numbered mbed pins can be used as an InterruptIn, except p19 and p20.
    InterruptIn pause(p30);                                //button to pause playing   S1                  
    InterruptIn real(p23);                                 //button realtime playing   S2                    
    InterruptIn seek(p22);                                 // seek                     S3
    InterruptIn mbutton(p21);                              // Menu                     S4 
    InterruptIn finish(p27);                               // on/off button            S5  
    DigitalInOut onoff(p27);
    
    InterruptIn pinA(p26);                                  //RPG p26
    DigitalIn   pinB(p24);    
                        
      
    DigitalOut led(LED1);                 

    //Timer t;
    Timeout after;
    
    bool menuEnd = true;
    int iValue = 0; //A variable that will be increased or decreased when we turn the encoder (RPG)
    int lastRPG = 0;
    
    // RDS data
    char Response[20];
    char PSN[9] = ("        ");
    char RT[65]=("                                                                ");
    
    /* Buffers, pointers and states */
    int circularRXBuffer[3072]; 
    int circularTXBuffer[3072]; 
    char buffer4[512]; 
    volatile bool recording = false;  
    volatile bool playing = true;  
    int sd_size;
    volatile int write_sector;
    volatile int read_sector;
    volatile int writeRXPointer;
    volatile int readRXPointer;
    volatile int writeTXPointer;
    volatile int readTXPointer;
    volatile int read4Pointer;
    volatile int thetaRX;
    volatile int thetaTX;
  
    /* Record settings */
    unsigned int frequency = 8000;                         //sample rate
    unsigned short wordWidth = 16;                          //bits per sample per channel
    unsigned short channels = 1;                            //number of channels
    // -> 60 blocks/sec

    /* Receiver settings */
    bool seekButton = false; // flag raised to send frequency to radio
    int temp;
    union sd {
        struct {
            int stations[7];                                   // frequency of pre-stored stations
            int lastStation;
            unsigned char Volume;
        } s;    
        char sd_buffer[512];
    }config;    
    
// Forward declarations
void startSeek(void);       
void pausePlaying(void);   
void realPlaying(void);   
void showRDS(void);      
void showRadio(void);       
void resetRecPlay(void);   
void showMenu(void);  
void getStation(void); 
void setStation(void); 
void menuReturn(void); 
void rewindPlaying(void); 
void encoderClick(void); 
/*-----------------------------------------------------------------------------------*/

/****************************************************************************
TLV320 part
*****************************************************************************/
/* Function to read content of I2SRXFIFO into circular buffer */
void record(void){ 
    audio.read();
    circularRXBuffer[writeRXPointer] = audio.rxBuffer[0];
    circularRXBuffer[++writeRXPointer] = audio.rxBuffer[1];
    circularRXBuffer[++writeRXPointer] = audio.rxBuffer[2];
    circularRXBuffer[++writeRXPointer] = audio.rxBuffer[3];
    ++writeRXPointer;
    thetaRX += 4;  
    if(writeRXPointer > 3070) writeRXPointer = 0;
}

/* Function to read from circular buffer and send data to TLV320 */
void play(void){ 
    audio.write(circularTXBuffer, readTXPointer, 8);
    //Pointer fun :-)
    readTXPointer += 8;
    if (readTXPointer >= 3072) readTXPointer = 0;
    if (thetaTX >= 8) thetaTX -= 8;
}

/* Function to select the interrupt source */
void record_play(void){ 
    int state;
    char rx_level;
    
    state = LPC_I2S->I2SSTATE;
    //printf("%08X ",state);
    rx_level = (state >> 8);
    if (rx_level >= 4) record(); else play();
}

/*Function to write data from circular buffer to sd storage */
void streamToSD(void){
    if(thetaRX > 512){                                                // only start writing when there is enough data!
        //sd.disk_write((const char *)(circularBuffer + readPointer), write_sector);  //write data in 512 byte chunks, this is
        sd.disk_write((const char *)&circularRXBuffer[readRXPointer], write_sector);  //write data in 512 byte chunks, this is
        //printf("%d %d\n",write_sector,read_sector);
        write_sector++;
        if (write_sector >= sd_size) write_sector = 1;
        if (thetaRX >= 128) thetaRX -= 128;                                               //the natural usb memory chunk size
        readRXPointer += 128;
        if(readRXPointer > 3070) readRXPointer = 0; 
    }
}

/* Function to stop recording, namely to stop audio interrupts and shutDown*/
void shutDown(void){
    recording = false;
    audio.stop();
    printf("W/R: %d %d\n",write_sector,read_sector);
    printf("readT/R: %d %d\n",readTXPointer,readRXPointer);
    printf("writeT/R: %d %d\n",writeTXPointer,writeRXPointer);
    printf("thetaR/T: %d %d\n",thetaRX,thetaTX);
    wait(0.5);
}

/* Function to get 4 integers from SD Card */
void disk_read4(char *fill_buffer) {
    if (read4Pointer >= 512) {
        sd.disk_read(buffer4, read_sector);
        read_sector++;
        if (read_sector >= sd_size) read_sector = 1;
        read4Pointer = 0;
        //printf("%d,%d ",buffer[0],buffer[1]);
    }
    memcpy(fill_buffer,&buffer4[read4Pointer],16);
    read4Pointer +=16;       
} 

/* Function to load circular buffer from SD Card */
void fillBuffer(void){
    if(thetaTX < 3072){
        disk_read4((char *)&circularTXBuffer[writeTXPointer]);   //read 4 integers into the circular buffer at a time
        //More pointer fun :D
        thetaTX+=4;
        writeTXPointer+=4;
        if (writeTXPointer >= 3072) writeTXPointer = 0;
    }
}

void resetRecPlay(void) {
    write_sector = 1;
    read_sector = 1;
    writeRXPointer = 0;
    readRXPointer = 0;
    writeTXPointer = 0;
    readTXPointer = 0;
    read4Pointer = 512;
    thetaRX = 0;
    thetaTX = 0;
    for(int j = 0; j < 3072; ++j){
         circularTXBuffer[j] = 0;              //clear circular buffer
         circularRXBuffer[j] = 0;              //clear circular buffer
    }
    
}

/****************************************************************************
Button S1 to S5 part
*****************************************************************************/
/* Associate buttons with functions */
void initButtons(void) {
    pause.mode(PullUp);         
    pause.fall(&pausePlaying);             //attach pause button S1
    real.mode(PullUp);         
    real.fall(&realPlaying);               //attach realtime button S2
    seek.mode(PullUp);         
    seek.fall(&startSeek);                 //attach seek button S3
    mbutton.mode(PullUp);         
    mbutton.fall(&showMenu);              // attach menu button S4     
    onoff.input();                        // S5  
    finish.mode(PullNone);         
    finish.fall(&shutDown);               //attach finish button S5
}

/* Function to enable pause button */
void enablePause(void){
    pause.fall(&pausePlaying);               //attach pause button    
}

/* Function to pause playing  S1 */
void pausePlaying(void){
    pause.fall(NULL);                              //disable pause button
    //finish.fall(NULL);                              //de-attach finish button
    if (playing) {
        playing = false;
        LPC_I2S->I2SIRQ = 0;
        audio.start(RECEIVE);                     //recording only
        printf("Pause:%d\n",read_sector);
        led = 1;
        lcd.cls();
        lcd.printf("%d secs",read_sector/35);
        //finish.fall(&rewindPlaying);           //attach finish button
    } else {
        playing = true;    
        LPC_I2S->I2SIRQ = 0;
        audio.start(BOTH);                      //recording and playing
        printf("Cont:%d\n",read_sector);
        led = 0;
        //finish.fall(&stopRecording);           //re-attach finish button
        showRadio();
   }
   after.attach(&enablePause,1.0);
}

/* Function to realtime playing  S2 */
void realPlaying(void){
    read_sector = write_sector;
    printf("Real:%d\n",read_sector);
    wait(0.1);    
}

/* Function to rewind playing */
void rewindPlaying(void){
    if (read_sector > 100) read_sector -= 100;   
    lcd.cls();
    lcd.printf("%d secs",read_sector/35);
    //printf("Rew:%d\n",read_sector); 
}
/* Function to forward playing */
void forwardPlaying(void){
    if ((read_sector + 100) < write_sector) read_sector += 100;   
    lcd.cls();
    lcd.printf("%d secs",read_sector/35);
    //printf("Fwd:%d\n",read_sector);
}

/* Menu for channel selection and frequency selection  S4 */
void showMenu(void) {
    //finish.fall(NULL);           //de-attach finish button
    pinA.rise(NULL);             //de-attach pinA from RPG interrupt function
    audio.stop();
    //printf("Menu\n");
    
    navigator.show();
    menuEnd = false;
    while(!menuEnd){
        navigator.poll(); // In a loop, call navigator's poll method to determine if the user is interacting with the RPG. 
    }    
    resetRecPlay();
    audio.start(BOTH);
    showRadio();
    //finish.fall(&stopRecording);           //attach finish button
    pinA.rise(&encoderClick);             //attach pinA to RPG interrupt function
    //printf("Menu End\n");
}

void dialFrequency(void) {
    int myFrequency;
    int a,b;
    
    //printf("dialFrequency\n");
    menuEnd = true;
    lcd.cls();
    myFrequency = radio.Frequency;
    
    while (!rpg.pb())  {
        lcd.locate(0,0);
        a = myFrequency/100;
        b = myFrequency - a*100;
        lcd.printf("%3d.%02d\n", a,b);
        if (rpg.dir() < 0) {
            if (myFrequency < 10800) myFrequency += 10;
        }   
        if (rpg.dir() > 0) {
            if (myFrequency > 8750) myFrequency -= 10;
        }    
    }
    radio.tuneFrequency(myFrequency);        // set selected frequency
}

/* Select a pre-stored radio station */
void getStation(void) {
    int x = 2;
    int a,b;
    int direction;  // 1 = CW, -1 = CCW
   
    //printf("getStaion\n");
    menuEnd = true;
    lcd.cls();
    lcd.printf("  1 2 3 4 5 6\n");
    lcd.Cursor(ON);
    
    while (!rpg.pb())  {
        a = config.s.stations[x/2]/100;
        b = config.s.stations[x/2] - a*100;
        lcd.locate(0,1);
        lcd.printf("%3d.%02d\n", a,b);
        lcd.setCursor(x,0);
        if((direction = rpg.dir())!=0){     //Get Dir
            wait(0.2); 
            if(direction > 0) {
                if (x > 2 ) x -=2;
            }
            if(direction < 0) {
                if (x < 12) x +=2;
            }    
        }
   }
   //printf("Station: %d\n",x/2);
   lcd.Cursor(OFF);
   radio.tuneFrequency(config.s.stations[x/2]);        // get selected frequency
   config.s.lastStation = x/2;
}

/* Assign radio stations frequency to station number */
void setStation(void) {
    int x = 2;
    int a,b;
    int direction;  // 1 = CW, -1 = CCW
   
    //printf("setStaion\n");
    menuEnd = true;
    lcd.cls();
    lcd.printf("  1 2 3 4 5 6\n");
    lcd.Cursor(ON);
    
    while (!rpg.pb())  {
        a = config.s.stations[x/2]/100;
        b = config.s.stations[x/2] - a*100;
        lcd.locate(0,1);
        lcd.printf("%3d.%02d\n", a,b);
        lcd.setCursor(x,0);
        if((direction = rpg.dir())!=0){     //Get Dir
            wait(0.2); 
            if(direction > 0) {
                if (x > 2 ) x -=2;
            }
            if(direction < 0) {
                if (x < 12) x +=2;
            }    
        }
   }
   //printf("Station: %d\n",x/2);
   lcd.Cursor(OFF);
   config.s.stations[x/2] = radio.Frequency;        // set selected frequency
   sd.disk_write(config.sd_buffer, 0);
}

void menuReturn(void) {
    printf("menuReturn\n");
    menuEnd = true;
}

/* Function to start seek  S3 */
void startSeek(void) {
    audio.stop();
    radio.seekUp();    
    while (!(radio.getIntStatus() & STCINT));
    while (!radio.getValidSignal());                
    //temp = radio.getFrequency();
    showRadio();
    printf("Current Station\n%5d MHz FM (%d,%d)\n", radio.Frequency,radio.RSSI,radio.SNR);
    resetRecPlay();
    audio.start(BOTH);
}

/****************************************************************************
* Si4735 part
* not used due to cpu performance limitation
*****************************************************************************/
void showRDS(void) {
    char GroupType;
    char ProgType;
    char Pos;
    char LtOffset,Sign;                 // Local time Offset, sign of LTO
    char Minute,Hour;
    //printf("rds\n");
        
    for (int i=0; i<20; i++) Response[i]=0;       
    radio.getFmRDSStatus(Response);
    
    while (Response[3] > 0 && (Response[12] & 3) == 0) {    // FIFO count && receive error
            GroupType = (Response[6] >> 4) & 15;
            ProgType  = (Response[7] >> 5);
            ProgType |= (Response[6] & 0x03);
            printf("RDS:%d %d\n",GroupType, ProgType);
        if (GroupType == 0) {                       // Basic tuning information
            Pos = (Response[7] & 3) * 2;
            PSN[Pos]=Response[10];
            PSN[Pos+1]=Response[11];
            PSN[8] = 0;                          // string termination
            printf("%s\n",PSN);
            printf("%c%c   %d\n",Response[10],Response[11],Pos);
        }
#ifdef rtx
        if (GroupType == 2) {                       // Radiotext
            printf("Radio Text----\n");
            Pos = (Response[7] & 0x0F) * 4;
            RT[Pos] = Response[8]; 
            RT[Pos+1] = Response[9]; 
            RT[Pos+2] = Response[10]; 
            RT[Pos+3] = Response[11]; 
            RT[64] = 0;
            printf("%s\n",RT);
        }            
#endif        
        if (GroupType == 4) {                       // assemble Time
            printf("Time----\n");
            LtOffset = (Response[11] & 0x0F) / 2;
            Sign = Response[11] & 0x10;
            Minute = (Response[11] >> 6) | ((Response[10] & 0x0F) << 2);
            Hour = (Response[10] >> 4) | ((Response[9] & 0x01) << 4);
            if (Sign == 0) Hour += LtOffset; else Hour -= LtOffset;
            printf("%02d:%02d\n",Hour,Minute);
        }

        for (int i=0; i<20; i++) Response[i]=0;       
        radio.getFmRDSStatus(Response);
    }  // end while
}


void showRadio(void) {
    int myFrequency;
    int a,b;
    
    lcd.cls();
    myFrequency = radio.getFrequency();
    a = myFrequency/100;
    b = myFrequency - a*100;
    lcd.printf("%3d.%02d (%d,%d)\n", a,b,radio.RSSI,radio.SNR);
/*    for (int i=0; i<100; i++) {
        showRDS();
        wait(0.01);
    }*/
}

void encoderClick(){
  // encoder must have turned one click because interrupt 0 was triggered
  // read value from both encoder pins
  int valA = pinA.read();
  int valB = pinB.read();
  
  // compare pins to determine in which direction encoder was turned
  if (valA != valB){
      // pinA just changed but pinB had not yet changed
      // Direction must be clockwise if A changes before B
      iValue++;
  }
  else{
      // pinA just changed and pinB had already done so.
      // Direction must be counter-clockwise if B changes before A
      iValue--;
  }
}

/******************** Main ****************************/
int main(){
    int result;
    unsigned char Vol;
    
    printf("TS_receiver2\n"); 
         
    lcd.cls();
    lcd.printf("Timeshift Radio\n");

    // Initialize menu part
    rootMenu.add(Selection(&getStation, 0, NULL, " Select Station"));
    rootMenu.add(Selection(&dialFrequency, 1, NULL, " Dial Frequency"));
    rootMenu.add(Selection(&setStation, 2, NULL, " Set Station"));   
    rootMenu.add(Selection(&menuReturn, 3, &rootMenu, " Exit Menu"));   

    // Init SD-disk
    result = sd.disk_initialize();
    if (result != 0) {
        printf("Init SD-Disk failed\n");
        exit(result);
    }    
    sd_size = sd.disk_sectors();
    printf("%d\n",sd_size);

    sd.disk_read(config.sd_buffer,0);
    if (config.s.stations[1] != 8760) {                 // first time ?
        for (int i=0; i<7; i++) config.s.stations[i] = 8760;
        config.s.Volume = 0x7F;
        config.s.lastStation = 1;
    }
    Vol = config.s.Volume;  

    initButtons();

    // Initialize radio part
    radio.Frequency = config.s.stations[config.s.lastStation];
    radio.begin(FM);
    showRadio();
        
    // Initialize time-shift part     
    audio.power(0x02);                      //power up all TLV320, just not microphone
    audio.inputVolume(0.8, 0.8);            //Set input volume 
    audio.outputVolume(&Vol);                
    audio.frequency(frequency);
    audio.format(wordWidth, (1-channels));  //note int mode = 1 - channels
    //audio.bypass(true);

    audio.attach(&record_play);             //attach record_play function to audio interrupt
    resetRecPlay();        
    recording = true;
    LPC_I2S->I2SIRQ = 0;
    audio.start(BOTH);                      //start recording and playing

// Init RPG        
    pinA.mode(PullUp);         
    pinB.mode(PullUp);         
    pinA.rise(&encoderClick);             //attach pinA to RPG interrupt function
        
    printf("Start RecPlay\n");
    //t.start();

    while(recording) {
        streamToSD();
        if ((read_sector < write_sector) && playing) fillBuffer();             //continually fill circular buffer
        if (playing) {                          // no pause
            //if (iValue > lastRPG) radio.volumeUp();   
            //if (iValue < lastRPG) radio.volumeDown();
            if (iValue > lastRPG) {Vol--; audio.outputVolume(&Vol);}                
            if (iValue < lastRPG) {Vol++; audio.outputVolume(&Vol);} 
        } else {
            if (iValue > lastRPG) rewindPlaying();   
            if (iValue < lastRPG) forwardPlaying();
        }             
        lastRPG = iValue;      
    }
    
    audio.stop();
    //t.stop();
    after.detach();
    config.s.Volume = Vol;
    sd.disk_write(config.sd_buffer,0);
    printf("Done\n"); 
    lcd.cls();
    lcd.printf("Shutting Down...\n"); 
    onoff.output();
    onoff = 0;          // power off
}
